
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>

#define mtrx(_,m,i,j) ((_)[(i)*(m) + (j)])

static double * read_matrix(FILE *f, unsigned *_n, unsigned *_m)
{
	unsigned i, j, n, m;
	double *r;

	if (fscanf(f, "%d %d\n", &n, &m) != 2)
		abort();
	r = malloc(sizeof(double) * n * m);

	for (i=0; i<n; i++)
		for (j=0; j<m; j++)
			if (fscanf(f, "%lf", &mtrx(r,m,i,j)) != 1)
				abort();

	*_n = n;
	*_m = m;
	return r;
}

static void print_matrix(double *r, unsigned n, unsigned m)
{
	unsigned i, j;
	for (i=0; i<n; i++) {
		for (j=0; j<m; j++)
			printf("\t%.4f", mtrx(r,m,i,j));
		printf("\n");
	}
}

static void cholesky_gaxpy(double *r, unsigned n, unsigned m)
{
/*
A(1:n,1) = A(1:n,1) / sqrt(A(1,1));
for j=2:n
	A(j:n,j) = A(j:n,j) - A(j:n,1:j-1) * A(j,1:j-1)'; % GAXPY-Operation
	A(j:n,j) = A(j:n,j) / sqrt(A(j,j));
end
*/

	unsigned i, j, k;
	double v, *w = malloc(sizeof(double) * n);

	v = sqrt(mtrx(r,m,0,0));
	for (i=0; i<n; i++)
		mtrx(r,m,i,0) /= v;

	for (j=1; j<m; j++) {
		for (i=j; i<n; i++) {
			v = 0.0;
			for (k=0; k<j; k++)
				v += mtrx(r,m,i,k) * mtrx(r,m,j,k);
			w[i] = v;
		}
		for (i=j; i<n; i++)
			mtrx(r,m,i,j) -= w[i];

		v = sqrt(mtrx(r,m,j,j));
		for (i=j; i<n; i++)
			mtrx(r,m,i,j) /= v;
	}

	for (i=0; i<n; i++)
		for (j=i+1; j<m; j++)
			mtrx(r,m,i,j) = 0;

	free(w);
}

int main(int argc, char **argv)
{
	struct timeval tv, tw;
	unsigned n, m;
	FILE *in = argc > 1 ? fopen(argv[1], "r") : stdin;
	double *r = read_matrix(in, &n, &m);

	fclose(in);

	gettimeofday(&tv, NULL);
	cholesky_gaxpy(r, n, m);
	gettimeofday(&tw, NULL);

	fprintf(stderr, "time: %f sec\n",
		tw.tv_sec - tv.tv_sec + (tw.tv_usec - tv.tv_usec) / 1e6);
	print_matrix(r, n, m);

	return 0;
}
